@@ -14,34 +14,34 @@ Plugin lifecycle:
     * Call Plugin.prerun() to generate new output
 """
 
+import glob
 import hashlib
 import importlib
 import importlib.metadata
 import logging
-import os.path
-from collections.abc import MutableMapping
-from typing import Dict, Iterable, List, Optional, Sequence, Union
+from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Type, Union
 
 import numpy as np
 import pandas as pd
 
-from countess.core.logger import Logger
 from countess.core.parameters import (
     ArrayParam,
     BaseParam,
     FileArrayParam,
     FileParam,
-    FileSaveParam,
+    HasSubParametersMixin,
     MultiParam,
     StringParam,
 )
 from countess.utils.pandas import get_all_columns
 from countess.utils.parallel import multiprocess_map
 
-PRERUN_ROW_LIMIT = 100000
+PRERUN_ROW_LIMIT: int = 100000
+
+logger = logging.getLogger(__name__)
 
 
-def get_plugin_classes():
+def get_plugin_classes() -> Iterable[Type["BasePlugin"]]:
     plugin_classes = set()
     try:
         # Python >= 3.10
@@ -65,14 +65,14 @@ def get_plugin_classes():
     return plugin_classes
 
 
-def load_plugin(module_name, class_name):
+def load_plugin(module_name: str, class_name: str, plugin_name: Optional[str] = None) -> "BasePlugin":
     module = importlib.import_module(module_name)
     plugin_class = getattr(module, class_name)
     assert issubclass(plugin_class, BasePlugin)
-    return plugin_class()
+    return plugin_class(plugin_name)
 
 
-class BasePlugin:
+class BasePlugin(HasSubParametersMixin):
     """Base class for all plugins.  Plugins exist as entrypoints, but also
     PluginManager checks that plugins subclass this class before accepting them
     as plugins."""
@@ -84,58 +84,22 @@ class BasePlugin:
     num_inputs: int = 1
     num_outputs: int = 1
 
-    parameters: MutableMapping[str, BaseParam] = {}
     show_preview: bool = True
 
     @property
     def version(self) -> str:
         raise NotImplementedError(f"{self.__class__}.version")
 
-    def __init__(self, plugin_name=None):
-        # Parameters store the actual values they are set to, so we copy them
-        # so that if the same plugin is used twice in a pipeline it will have
-        # its own parameters.
+    def __init__(self, plugin_name: Optional[str] = None):
+        super().__init__()
 
         if plugin_name is not None:
             self.name = plugin_name
 
-        self.parameters = dict((k, v.copy()) for k, v in self.parameters.items())
-
-        # XXX should we allow django-esque declarations like this?
-        # Code gets cleaner, Namespace gets cluttered, though.
-
-        for key in dir(self):
-            if isinstance(getattr(self, key), BaseParam):
-                self.parameters[key] = getattr(self, key).copy()
-                setattr(self, key, self.parameters[key])
-
-    def add_parameter(self, name: str, param: BaseParam):
-        self.parameters[name] = param.copy()
-        return self.parameters[name]
-
-    def set_parameter(self, key: str, value: Union[bool, int, float, str], base_dir: str = "."):
-        param = self.parameters
-        for k in key.split("."):
-            if k == "_label" and hasattr(param, "label"):
-                param.label = value
-                return
-
-            # XXX types are a mess here
-            param = param[k]  # type: ignore
-        if isinstance(param, (FileParam, FileSaveParam)) and value is not None:
-            assert isinstance(value, str)
-            param.value = os.path.join(base_dir, value)
-        else:
-            param.value = value  # type: ignore
-
-    def get_parameters(self, base_dir="."):
-        for key, parameter in self.parameters.items():
-            yield from parameter.get_parameters(key, base_dir)
-
     def get_parameter_hash(self):
         """Build a hash of all configuration parameters"""
         h = hashlib.sha256()
-        for k, v in self.parameters.items():
+        for k, v in self.params.items():
             h.update((k + "\0" + v.get_hash_value()).encode("utf-8"))
         return h
 
@@ -152,17 +116,21 @@ class ProcessPlugin(BasePlugin):
     def prepare(self, sources: List[str], row_limit: Optional[int] = None):
         pass
 
-    def process(self, data, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
+    def preprocess(self, data, source: str) -> None:
+        """Called with each `data` input from `source` before `process` is called
+        for that data, to set up config etc.  Can't return anything."""
+
+    def process(self, data, source: str) -> Iterable[pd.DataFrame]:
         """Called with each `data` input from `source`, yields results"""
         raise NotImplementedError(f"{self.__class__}.process")
 
-    def finished(self, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
+    def finished(self, source: str) -> Iterable[pd.DataFrame]:
         """Called when a `source` is finished and not able to
         send any more messages.  Can be ignored by most things."""
         # override this if you need to do anything
         return []
 
-    def finalize(self, logger: Logger) -> Iterable[pd.DataFrame]:
+    def finalize(self) -> Iterable[pd.DataFrame]:
         """Called when all sources are finished.  Can be
         ignored by most things.  This should reset the
         plugin to be ready for another use."""
@@ -171,93 +139,136 @@ class ProcessPlugin(BasePlugin):
 
 
 class SimplePlugin(ProcessPlugin):
-    def process(self, data, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
+    def process(self, data, source: str) -> Iterable[pd.DataFrame]:
         """Called with each `data` input from `source`, calls
         `callback` to send messages to the next plugin"""
         raise NotImplementedError(f"{self.__class__}.process")
 
 
-class FileInputPlugin(BasePlugin):
-    """Mixin class to indicate that this plugin can read files from local
-    storage."""
+class FileInputPluginFilesMultiParam(MultiParam):
+    """MultiParam which only asks for one thing, a filename.  Using a MultiParam wrapper
+    because that makes it easier to extend the FileInputPlugin."""
 
-    file_number = 0
-    name = ""
-    row_limit = None
+    filename = FileParam("Filename")
+
+
+class FileInputPlugin(BasePlugin):
+    files = FileArrayParam("Files", FileInputPluginFilesMultiParam("File"))
+    file_types: Sequence[tuple[str, Union[str, list[str]]]] = [("Any", "*")]
+    row_limit: Optional[int] = None
     num_inputs = 0
 
-    # used by the GUI file dialog
-    file_types: Sequence[tuple[str, Union[str, list[str]]]] = [("Any", "*")]
-    file_params: MutableMapping[str, BaseParam] = {}
+    def __init__(self, *a, **k):
+        # Add in filenames
+        super().__init__(*a, **k)
+        self.files.file_types = self.file_types
 
-    def num_files(self) -> int:
-        """return the number of 'files' which are to be loaded.  The pipeline
-        will call code equivalent to
-        `[ p.load_file(n, logger, row_limit) for n in range(0, p.num_files() ]`
-        although potentially using threads, multiprocessing, etc."""
-        raise NotImplementedError("FileInputMixin.num_files")
+    def filenames_and_params(self):
+        for file_param in self.files:
+            for filename in glob.iglob(file_param.filename.value):
+                yield filename, file_param
 
-    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable:
-        """Called potentially from multiple processes, see FileInputMixin.num_files()"""
-        raise NotImplementedError("FileInputMixin.load_file")
+    def read_file_to_dataframe(self, filename: str, file_param: BaseParam, row_limit=None) -> Any:
+        """May be called from multiple processes at once.  Note that the 'filename' parameter
+        overrides the 'file_param.filename.value' as the latter may be a glob."""
+        raise NotImplementedError("{self.class}.read_file_to_dataframe")
+
+    def load_file(self, filename_and_param: Tuple[str, BaseParam], row_limit: Optional[int] = None) -> Iterable:
+        filename, file_param = filename_and_param
+        yield self.read_file_to_dataframe(filename, file_param, row_limit)
 
     def prepare(self, sources: List[str], row_limit: Optional[int] = None):
+        assert len(sources) == 0
         self.row_limit = row_limit
 
-    def finalize(self, logger: Logger) -> Iterable:
-        num_files = self.num_files()
+    def finalize(self) -> Iterable:
+        filenames_and_params = list(self.filenames_and_params())
+        num_files = len(filenames_and_params)
+        logger.info("%s: 0%%", self.name)
         if num_files > 1:
             row_limit_per_file = self.row_limit // num_files if self.row_limit else None
-            yield from multiprocess_map(self.load_file, range(0, num_files), logger, row_limit_per_file)
+            for n, r in enumerate(multiprocess_map(self.load_file, filenames_and_params, row_limit_per_file)):
+                if n < num_files:
+                    logger.info("%s: %d%%", self.name, int(100 * n / num_files))
+                yield r
         elif num_files == 1:
-            yield from self.load_file(0, logger, self.row_limit)
+            yield from self.load_file(filenames_and_params[0], self.row_limit)
+        logger.info("%s: 100%%", self.name)
 
 
 class PandasProcessPlugin(ProcessPlugin):
     DATAFRAME_BUFFER_SIZE = 100000
 
-    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
-        raise NotImplementedError(f"{self.__class__}.process")
-
-
-class PandasSimplePlugin(SimplePlugin):
-    """Base class for plugins which accept and return pandas DataFrames.
-    Subclassing this hides all the distracting aspects of the pipeline
-    from the plugin implementor, who only needs to override process_dataframe"""
-
     input_columns: Dict[str, np.dtype] = {}
 
     def prepare(self, sources: list[str], row_limit: Optional[int] = None):
         self.input_columns = {}
 
-    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
+    def preprocess(self, data: pd.DataFrame, source: str) -> None:
+        self.input_columns.update(get_all_columns(data))
+
+    def process(self, data: pd.DataFrame, source: str) -> Iterable[pd.DataFrame]:
+        raise NotImplementedError(f"{self.__class__}.process")
+
+    def finalize(self) -> Iterable[pd.DataFrame]:
+        yield from super().finalize()
+        self.set_column_choices(self.input_columns.keys())
+
+
+class PandasConcatProcessPlugin(PandasProcessPlugin):
+    # Like PandsaProcessPlugin but collect all the inputs together before trying to do anything
+    # with them.
+
+    def __init__(self, *a, **k) -> None:
+        super().__init__(*a, **k)
+        self.dataframes: list[pd.DataFrame] = []
+
+    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
+        super().prepare(sources, row_limit)
+        self.dataframes = []
+
+    def process(self, data: pd.DataFrame, source: str) -> Iterable:
+        self.dataframes.append(data)
+        return []
+
+    def finalize(self) -> Iterable[pd.DataFrame]:
+        data_in = pd.concat(self.dataframes)
+        data_out = self.process_dataframe(data_in)
+        if data_out is not None:
+            yield data_out
+        yield from super().finalize()
+
+    def process_dataframe(self, dataframe: pd.DataFrame) -> Optional[pd.DataFrame]:
+        """Override this to process a single dataframe"""
+        raise NotImplementedError(f"{self.__class__}.process_dataframe()")
+
+
+class PandasSimplePlugin(PandasProcessPlugin):
+    """Base class for plugins which accept and return pandas DataFrames.
+    Subclassing this hides all the distracting aspects of the pipeline
+    from the plugin implementor, who only needs to override process_dataframe"""
+
+    def process(self, data: pd.DataFrame, source: str) -> Iterable[pd.DataFrame]:
         """Just deal with each dataframe as it comes.  PandasSimplePlugins don't care about `source`."""
         assert isinstance(data, pd.DataFrame)
 
-        self.input_columns.update(get_all_columns(data))
-
         try:
-            result = self.process_dataframe(data, logger)
+            result = self.process_dataframe(data)
             if result is not None:
                 assert isinstance(result, pd.DataFrame)
                 if len(result) > 0:
                     yield result
 
         except Exception as exc:  # pylint: disable=broad-exception-caught
-            logger.exception(exc)
+            logger.warning("Exception", exc_info=exc)  # pragma: no cover
 
-    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> Optional[pd.DataFrame]:
+    def process_dataframe(self, dataframe: pd.DataFrame) -> Optional[pd.DataFrame]:
         """Override this to process a single dataframe"""
         raise NotImplementedError(f"{self.__class__}.process_dataframe()")
 
-    def finalize(self, logger: Logger) -> Iterable[pd.DataFrame]:
-        yield from super().finalize(logger)
-        for p in self.parameters.values():
-            p.set_column_choices(self.input_columns.keys())
-
 
 # class MapReduceFinalizePlugin(BasePlugin):
-#    def map(self, data, logger: Logger) -> Iterable:
+#    def map(self, data) -> Iterable:
 #        return []
 #
 #    def reduce(self, data: Iterable):
@@ -268,7 +279,7 @@ class PandasSimplePlugin(SimplePlugin):
 #
 #
 # class PandasMapReduceFinalizePlugin(MapReduceFinalizePlugin):
-#    def map(self, data, logger: Logger) -> Iterable[pd.DataFrame]:
+#    def map(self, data) -> Iterable[pd.DataFrame]:
 #        return []
 #
 #    def reduce(self, data: Iterable[pd.DataFrame]) -> pd.DataFrame:
@@ -303,13 +314,13 @@ class PandasProductPlugin(PandasProcessPlugin):
         self.mem1 = []
         self.mem2 = []
 
-    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
+    def process(self, data: pd.DataFrame, source: str) -> Iterable[pd.DataFrame]:
         if source == self.source1:
             if self.mem1 is not None:
                 self.mem1.append(data)
             assert self.mem2 is not None
             for val2 in self.mem2:
-                df = self.process_dataframes(data, val2, logger)
+                df = self.process_dataframes(data, val2)
                 if len(df):
                     yield df
 
@@ -318,14 +329,14 @@ class PandasProductPlugin(PandasProcessPlugin):
                 self.mem2.append(data)
             assert self.mem1 is not None
             for val1 in self.mem1:
-                df = self.process_dataframes(val1, data, logger)
+                df = self.process_dataframes(val1, data)
                 if len(df):
                     yield df
 
         else:
             raise ValueError(f"Unknown source {source}")
 
-    def finished(self, source: str, logger: Logger) -> Iterable:
+    def finished(self, source: str) -> Iterable:
         if source == self.source1:
             # source1 is finished, mem2 is no longer needed
             self.source1 = None
@@ -338,13 +349,13 @@ class PandasProductPlugin(PandasProcessPlugin):
             raise ValueError(f"Unknown source {source}")
         return []
 
-    def finalize(self, logger: Logger) -> Iterable:
+    def finalize(self) -> Iterable:
         # free up any memory taken up by memoization
         self.mem1 = None
         self.mem2 = None
         return []
 
-    def process_dataframes(self, dataframe1: pd.DataFrame, dataframe2: pd.DataFrame, logger: Logger) -> pd.DataFrame:
+    def process_dataframes(self, dataframe1: pd.DataFrame, dataframe2: pd.DataFrame) -> pd.DataFrame:
         raise NotImplementedError(f"{self.__class__}.process_dataframes")
 
 
@@ -378,10 +389,11 @@ class PandasTransformBasePlugin(PandasSimplePlugin):
     def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
         raise NotImplementedError(f"{self.__class__}.series_to_dataframe()")
 
-    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
+    def dataframe_to_series(self, dataframe: pd.DataFrame) -> pd.Series:
         raise NotImplementedError(f"{self.__class__}.dataframe_to_series()")
 
-    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> Optional[pd.DataFrame]:
+    def process_dataframe(self, dataframe: pd.DataFrame) -> Optional[pd.DataFrame]:
+        dataframe_merged = None
         try:
             # 1. A dataframe with duplicates in its index can't be merged back correctly
             # in Step 4, so we add in an extra RangeIndex to guarantee uniqueness,
@@ -393,7 +405,7 @@ class PandasTransformBasePlugin(PandasSimplePlugin):
             # a series of results by PandasTransform...ToXMixin.dataframe_to_series,
             # which is expected to take each row of the dataframe, do something with
             # it and return a series of objects (values, tuples or dicts).
-            series = self.dataframe_to_series(dataframe, logger)
+            series = self.dataframe_to_series(dataframe)
 
             # 3. the series is expanded back out into rows by
             # PandasTransformXTo...Mixin.series_to_dataframe()
@@ -406,9 +418,8 @@ class PandasTransformBasePlugin(PandasSimplePlugin):
             if "__tmpidx" in dataframe_merged.index.names:
                 dataframe_merged.reset_index("__tmpidx", drop=True, inplace=True)
 
-        except Exception as exc:  # pylint: disable=broad-exception-caught
-            logger.exception(exc)
-            return None
+        except Exception as exc:  # pylint: disable=broad-exception-caught  # pragma: no cover
+            logger.warning("Exception", exc_info=exc)  # pragma: no cover
 
         return dataframe_merged
 
@@ -421,19 +432,19 @@ class PandasTransformSingleToXMixin:  # type: ignore [attr-defined]
     """Transformer which takes a single column, the name of which is specified
     in a ColumnChoiceParam called "column" """
 
-    def process_value(self, value, logger: Logger):
+    def process_value(self, value):
         raise NotImplementedError(f"{self.__class__}.process_value()")
 
-    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
-        column_name = self.parameters["column"].value  # type: ignore [attr-defined]
+    def dataframe_to_series(self, dataframe: pd.DataFrame) -> pd.Series:
+        column_name = self.params["column"].value  # type: ignore [attr-defined]
         if column_name in dataframe:
-            return dataframe[column_name].apply(self.process_value, logger=logger)
+            return dataframe[column_name].apply(self.process_value)
         elif column_name == dataframe.index.name:
-            return dataframe.index.to_series().apply(self.process_value, logger=logger)
+            return dataframe.index.to_series().apply(self.process_value)
         elif column_name in dataframe.index.names:
-            return dataframe.index.to_frame()[column_name].apply(self.process_value, logger=logger)
+            return dataframe.index.to_frame()[column_name].apply(self.process_value)
         else:
-            null_values = [self.process_value(None, logger)] * len(dataframe)
+            null_values = [self.process_value(None)] * len(dataframe)
             s = pd.Series(null_values, index=dataframe.index)
             return s
 
@@ -442,17 +453,17 @@ class PandasTransformRowToXMixin:
     """Transformer which takes an entire row. Less efficient than processing
     just a single value, but a bit more flexible."""
 
-    def process_row(self, row: pd.Series, logger: Logger):
+    def process_row(self, row: pd.Series):
         raise NotImplementedError(f"{self.__class__}.process_row()")
 
-    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
-        return dataframe.apply(self.process_row, axis=1, args=(logger,))
+    def dataframe_to_series(self, dataframe: pd.DataFrame) -> pd.Series:
+        return dataframe.apply(self.process_row, axis=1)
 
 
 class PandasTransformDictToXMixin:
     """Transformer which takes a row as a dictionary"""
 
-    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
+    def dataframe_to_series(self, dataframe: pd.DataFrame) -> pd.Series:
         if dataframe.index.names == [None]:
             # unnamed index,
             # XXX there is a bug in Pandas 2.1.x which prevents
@@ -461,7 +472,7 @@ class PandasTransformDictToXMixin:
             # hopefully this lambda can be removed some day.
             # https://github.com/pandas-dev/pandas/issues/55009
             return dataframe.apply(
-                lambda x: self.process_raw(x, list(dataframe.columns), logger),
+                lambda x: self.process_raw(x, list(dataframe.columns)),
                 axis=1,
                 raw=True,
             )
@@ -471,23 +482,23 @@ class PandasTransformDictToXMixin:
         if len(dataframe.index.names) == 1:
             # single index
             values = (
-                self.process_raw(list(index_and_data_tuple), columns, logger)
+                self.process_raw(list(index_and_data_tuple), columns)
                 for index_and_data_tuple in dataframe.itertuples(name=None)
             )
         else:
             # multiindex
             values = (
-                self.process_raw(list(index_tuple) + list(data_tuple), columns, logger)
+                self.process_raw(list(index_tuple) + list(data_tuple), columns)
                 for index_tuple, *data_tuple in dataframe.itertuples(name=None)
             )
 
         return pd.Series(values, index=dataframe.index)
 
-    def process_dict(self, data, logger: Logger):
+    def process_dict(self, data):
         raise NotImplementedError(f"{self.__class__}.process_dict()")
 
-    def process_raw(self, data, columns, logger: Logger) -> pd.Series:
-        return self.process_dict(dict(zip(columns, data)), logger=logger)
+    def process_raw(self, data, columns) -> pd.Series:
+        return self.process_dict(dict(zip(columns, data)))
 
 
 class PandasTransformXToSingleMixin:
@@ -496,10 +507,10 @@ class PandasTransformXToSingleMixin:
 
     def __init__(self, *a, **k):
         super().__init__(*a, **k)
-        assert isinstance(self.parameters["output"], StringParam)
+        assert isinstance(self.params["output"], StringParam)
 
     def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
-        return series.to_frame(name=self.parameters["output"].value)  # type: ignore [attr-defined]
+        return series.to_frame(name=self.params["output"].value)  # type: ignore [attr-defined]
 
 
 class PandasTransformXToTupleMixin:
@@ -508,27 +519,23 @@ class PandasTransformXToTupleMixin:
 
     def __init__(self, *a, **k):
         super().__init__(*a, **k)
-        assert hasattr(self, "parameters")
-        assert isinstance(self.parameters["output"], ArrayParam)
-        assert all(isinstance(pp["name"], StringParam) for pp in self.parameters["output"])
+        assert hasattr(self, "params")
+        assert isinstance(self.params["output"], ArrayParam)
+        assert all(isinstance(pp["name"], StringParam) for pp in self.params["output"])
 
     def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
-        assert hasattr(self, "parameters")
-        assert isinstance(self.parameters["output"], ArrayParam)
-        assert all(isinstance(pp["name"], StringParam) for pp in self.parameters["output"])
+        assert hasattr(self, "params")
+        assert isinstance(self.params["output"], ArrayParam)
+        assert all(isinstance(pp["name"], StringParam) for pp in self.params["output"])
         column_names = [
-            pp.name.value or "Column %d" % n for n, pp in enumerate(self.parameters["output"], 1)
+            pp.name.value or "Column %d" % n for n, pp in enumerate(self.params["output"], 1)
         ]  # type: ignore [attr-defined]
 
         series.dropna(inplace=True)
         data = series.tolist()
-        if len(data):
-            max_cols = max(len(d) for d in data)
-            column_names = column_names[:max_cols]
-            df = pd.DataFrame(data, columns=column_names, index=series.index)
-            return df
-        else:
-            return pd.DataFrame()
+        max_cols = max(len(d) for d in data) if len(data) else 0
+        column_names = column_names[:max_cols]
+        return pd.DataFrame(data, columns=column_names, index=series.index)
 
 
 class PandasTransformXToDictMixin:
@@ -548,7 +555,7 @@ class PandasTransformSingleToSinglePlugin(
 ):
     """Transformer which takes a single column and returns a single value"""
 
-    def process_value(self, value, logger: Logger):
+    def process_value(self, value):
         raise NotImplementedError(f"{self.__class__}.process_value()")
 
 
@@ -557,7 +564,7 @@ class PandasTransformSingleToTuplePlugin(
 ):
     """Transformer which takes a single column and returns a tuple of values"""
 
-    def process_value(self, value, logger: Logger) -> Optional[Iterable]:
+    def process_value(self, value) -> Optional[Iterable]:
         raise NotImplementedError(f"{self.__class__}.process_value()")
 
 
@@ -566,7 +573,7 @@ class PandasTransformSingleToDictPlugin(
 ):
     """Transformer which takes a single column and returns a dictionary of values"""
 
-    def process_value(self, value, logger: Logger) -> Optional[Dict]:
+    def process_value(self, value) -> Optional[Dict]:
         raise NotImplementedError(f"{self.__class__}.process_value()")
 
 
@@ -575,7 +582,7 @@ class PandasTransformRowToSinglePlugin(
 ):
     """Transformer which takes a whole row and returns a single value"""
 
-    def process_row(self, row: pd.Series, logger: Logger):
+    def process_row(self, row: pd.Series):
         raise NotImplementedError(f"{self.__class__}.process_row()")
 
 
@@ -584,7 +591,7 @@ class PandasTransformRowToTuplePlugin(
 ):
     """Transformer which takes a whole row and returns a tuple of values"""
 
-    def process_row(self, row: pd.Series, logger: Logger):
+    def process_row(self, row: pd.Series):
         raise NotImplementedError(f"{self.__class__}.process_row()")
 
 
@@ -593,7 +600,7 @@ class PandasTransformRowToDictPlugin(
 ):
     """Transformer which takes a whole row and returns a dictionary of values"""
 
-    def process_row(self, row: pd.Series, logger: Logger):
+    def process_row(self, row: pd.Series):
         raise NotImplementedError(f"{self.__class__}.process_row()")
 
 
@@ -602,7 +609,7 @@ class PandasTransformDictToSinglePlugin(
 ):
     """Transformer which takes a whole row and returns a single value"""
 
-    def process_dict(self, data, logger: Logger):
+    def process_dict(self, data):
         raise NotImplementedError(f"{self.__class__}.process_dict()")
 
 
@@ -611,7 +618,7 @@ class PandasTransformDictToTuplePlugin(
 ):
     """Transformer which takes a whole row and returns a tuple of values"""
 
-    def process_dict(self, data, logger: Logger):
+    def process_dict(self, data):
         raise NotImplementedError(f"{self.__class__}.process_dict()")
 
 
@@ -620,46 +627,29 @@ class PandasTransformDictToDictPlugin(
 ):
     """Transformer which takes a whole row and returns a dictionary of values"""
 
-    def process_dict(self, data, logger: Logger):
+    def process_dict(self, data):
         raise NotImplementedError(f"{self.__class__}.process_dict()")
 
 
-class PandasInputPlugin(FileInputPlugin):
-    """A specialization of the PandasBasePlugin to allow it to follow nothing,
-    eg: come first."""
-
-    def num_files(self):
-        raise NotImplementedError(f"{self.__class__}.num_files()")
-
-    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable[pd.DataFrame]:
-        raise NotImplementedError(f"{self.__class__}.load_file()")
+class PandasInputFilesPlugin(FileInputPlugin):
+    def read_file_to_dataframe(self, filename: str, file_param: BaseParam, row_limit=None) -> pd.DataFrame:
+        """May be called from multiple processes at once.  Note that the 'filename' parameter
+        overrides the 'file_param.filename.value' as the latter may be a glob."""
+        raise NotImplementedError("{self.class}.read_file_to_dataframe")
 
 
-class PandasInputFilesPlugin(PandasInputPlugin):
-    def __init__(self, *a, **k):
-        # Add in filenames
-        super().__init__(*a, **k)
-        file_params = {"filename": FileParam("Filename", file_types=self.file_types)}
-        file_params.update(self.file_params)
+class PandasInputPlugin(PandasProcessPlugin):
+    num_inputs = 0
 
-        self.parameters = dict(
-            [("files", FileArrayParam("Files", MultiParam("File", file_params)))] + list(self.parameters.items())
-        )
+    def process(self, data: pd.DataFrame, source: str) -> Iterable[pd.DataFrame]:
+        return []
 
-    def num_files(self):
-        return len(self.parameters["files"].params)
-
-    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable[pd.DataFrame]:
-        assert isinstance(self.parameters["files"], ArrayParam)
-        file_params = self.parameters["files"][file_number]
-        yield self.read_file_to_dataframe(file_params, logger, row_limit)
-
-    def read_file_to_dataframe(self, file_params, logger, row_limit=None) -> pd.DataFrame:
-        raise NotImplementedError(f"{self.__class__}.read_file_to_dataframe")
+    def finalize(self) -> Iterable[pd.DataFrame]:
+        raise NotImplementedError("{self.class}.finalize")
 
 
 class PandasOutputPlugin(PandasProcessPlugin):
     num_outputs = 0
 
-    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
+    def process(self, data: pd.DataFrame, source: str) -> Iterable[pd.DataFrame]:
         raise NotImplementedError(f"{self.__class__}.process")
