@@ -13,11 +13,13 @@ from ..core.parameters import (
     BooleanParam,
     ChoiceParam,
     ColumnOrStringParam,
+    DictChoiceParam,
     FileArrayParam,
     FileParam,
     FileSaveParam,
+    FramedMultiParam,
     MultiParam,
-    SimpleParam,
+    ScalarParam,
     TabularMultiParam,
     TextParam,
 )
@@ -59,22 +61,22 @@ class ParameterWrapper:
         self.subwrapper_buttons: list[tk.Button] = []
         self.column_labels: list[tk.Label] = []
 
-        self.subwrappers: Mapping[BaseParam, ParameterWrapper] = {}
+        self.subwrappers: Mapping[int, ParameterWrapper] = {}
 
         self.var: Optional[tk.Variable] = None
         self.entry: Optional[tk.Widget] = None
         self.label: Optional[tk.Widget] = None
         self.row_labels: list[tk.Widget] = []
 
-        if isinstance(parameter, ArrayParam):
+        if isinstance(parameter, (ArrayParam, FramedMultiParam)):
             self.label = None
         else:
             self.label = tk.Label(tk_parent, text=parameter.label)
 
-        if isinstance(parameter, ChoiceParam):
+        if isinstance(parameter, (ChoiceParam, DictChoiceParam)):
             self.var = tk.StringVar(tk_parent, value=parameter.value)
             self.entry = ttk.Combobox(tk_parent, textvariable=self.var)
-            self.entry["values"] = parameter.choices or [""]
+            self.entry["values"] = parameter.get_values() or [""]
             if isinstance(parameter, ColumnOrStringParam):
                 self.entry.bind("<Key>", self.combobox_set)
                 self.entry["state"] = "normal"
@@ -91,12 +93,12 @@ class ParameterWrapper:
             # so we use a different method!
             # XXX is this a simpler way to handle other kinds of fields too?
             self.entry = tk.Text(tk_parent, height=10)
-            self.entry.insert("1.0", parameter.value)
+            self.entry.insert("1.0", parameter.value or "")
             if parameter.read_only:
                 self.entry["state"] = tk.DISABLED
             else:
                 self.entry.bind("<<Modified>>", self.widget_modified_callback)
-        elif isinstance(parameter, SimpleParam):
+        elif isinstance(parameter, ScalarParam):
             self.var = tk.StringVar(tk_parent, value=parameter.value)
             self.entry = tk.Entry(tk_parent, textvariable=self.var)
             if parameter.read_only:
@@ -115,7 +117,7 @@ class ParameterWrapper:
                     text=f"Add {parameter.param.label}",
                     command=self.add_row_callback,
                 )
-                self.button.grid(row=len(parameter.params) + 1)
+                self.button.grid(row=len(parameter.params), padx=10)
 
         elif isinstance(parameter, ArrayParam) and (self.level < 3 or isinstance(parameter.param, TabularMultiParam)):
             label_frame_label = tk.Frame(tk_parent)
@@ -139,6 +141,15 @@ class ParameterWrapper:
 
                 self.update_subwrappers(parameter.params, drc)
 
+        elif isinstance(parameter, FramedMultiParam):
+            label_frame_label = tk.Frame(tk_parent)
+            tk.Label(label_frame_label, text=parameter.label).grid(row=0, column=0, padx=5)
+            self.entry = tk.LabelFrame(tk_parent, labelwidget=label_frame_label, padx=10, pady=5)
+            self.entry.columnconfigure(0, weight=0)
+            self.entry.columnconfigure(1, weight=0)
+            self.entry.columnconfigure(2, weight=1)
+            self.update_subwrappers(parameter.params.values(), None)
+
         elif isinstance(parameter, (ArrayParam, MultiParam)):
             self.entry = tk.Frame(tk_parent)
             self.entry.columnconfigure(0, weight=0)
@@ -175,8 +186,6 @@ class ParameterWrapper:
             and not isinstance(self.parameter.param, TabularMultiParam)
         ):
             self.update_subwrappers_framed(self.parameter.params, self.delete_row_callback)
-            if self.button:
-                self.button.grid(row=len(self.parameter.params) + 1, padx=10)
         elif isinstance(self.parameter, ArrayParam) and isinstance(self.parameter.param, MultiParam) and self.level < 3:
             self.update_subwrappers_tabular(
                 self.parameter.params,
@@ -195,45 +204,47 @@ class ParameterWrapper:
                 )
         elif isinstance(self.parameter, MultiParam):
             self.update_subwrappers(self.parameter.params.values(), None)
-        elif isinstance(self.parameter, ChoiceParam):
-            choices = self.parameter.choices or [""]
-            self.entry["values"] = choices
+        elif isinstance(self.parameter, (ChoiceParam, DictChoiceParam)):
+            self.entry["values"] = self.parameter.get_values() or [""]
             self.var.set(self.parameter.value)
         elif isinstance(self.parameter, BooleanParam):
             self.set_checkbox_value()
         elif isinstance(self.parameter, TextParam):
             if self.parameter.read_only:
                 self.entry["state"] = "normal"
-            self.entry.replace("1.0", tk.END, self.parameter.value)
+            self.entry.replace("1.0", tk.END, self.parameter.value or "")
             if self.parameter.read_only:
                 self.entry["state"] = "disabled"
         elif isinstance(self.parameter, FileParam):
             self.entry["text"] = self.parameter.value
         else:
-            self.var.set(self.parameter.value)
+            if self.parameter.value is None:
+                self.var.set("")
+            else:
+                self.var.set(self.parameter.value)
 
         if self.label:
             self.label["text"] = self.parameter.label
 
     def cull_subwrappers(self, params):
-        params_set = set(params)
-        for p, pw in self.subwrappers.items():
-            if p not in params_set:
+        params_set = set(id(p) for p in params)
+        for pid, pw in self.subwrappers.items():
+            if pid not in params_set:
                 pw.destroy()
 
     def update_subwrappers(self, params, delete_row_callback):
         for n, p in enumerate(params):
-            if p in self.subwrappers:
-                self.subwrappers[p].update()
+            if id(p) in self.subwrappers:
+                self.subwrappers[id(p)].update()
             else:
-                self.subwrappers[p] = ParameterWrapper(
+                self.subwrappers[id(p)] = ParameterWrapper(
                     self.entry,
                     p,
                     self.callback,
                     delete_row_callback,
                     level=self.level + 1,
                 )
-            self.subwrappers[p].set_row(n)
+            self.subwrappers[id(p)].set_row(n)
 
         self.cull_subwrappers(params)
 
@@ -245,6 +256,9 @@ class ParameterWrapper:
         while self.column_labels:
             self.column_labels.pop().destroy()
 
+        # `MultiParam.param` is the 'template' each row in `MultiParam.params`
+        # is copied from ... step through its subparameters and make them
+        # each a column heading.
         for n, pp in enumerate(self.parameter.param.values()):
             column_label = tk.Label(self.entry, text=pp.label)
             self.column_labels.append(column_label)
@@ -258,17 +272,17 @@ class ParameterWrapper:
 
             subparams = p.params.values()
             for m, pp in enumerate(subparams):
-                if pp in self.subwrappers:
-                    self.subwrappers[pp].update()
+                if id(pp) in self.subwrappers:
+                    self.subwrappers[id(pp)].update()
                 else:
-                    self.subwrappers[pp] = ParameterWrapper(
+                    self.subwrappers[id(pp)] = ParameterWrapper(
                         self.entry,
                         pp,
                         self.callback,
                         delete_row_callback,
                         level=self.level + 1,
                     )
-                self.subwrappers[pp].entry.grid(row=n + 1, column=m + 1, padx=10)
+                self.subwrappers[id(pp)].entry.grid(row=n + 1, column=m + 1, padx=10)
             if not self.parameter.read_only and delete_row_callback:
                 button = delete_button(self.entry, command=partial(delete_row_callback, self, n))
                 button.grid(row=n + 1, column=len(subparams) + 1, padx=10)
@@ -278,20 +292,21 @@ class ParameterWrapper:
 
     def update_subwrappers_framed(self, params, delete_row_callback):
         for n, p in enumerate(params):
-            if p in self.subwrappers:
-                self.subwrappers[p].update()
+            if id(p) in self.subwrappers:
+                self.subwrappers[id(p)].update()
+                self.subwrappers[id(p)].entry.master.grid(row=n)
             else:
                 label_frame_label = tk.Frame(self.entry)
                 tk.Label(label_frame_label, text=p.label).grid(row=0, column=0, padx=10)
                 label_frame = tk.LabelFrame(self.entry, labelwidget=label_frame_label, padx=10)
-                label_frame.grid(sticky=tk.NSEW, padx=10)
+                label_frame.grid(row=n, sticky=tk.EW, padx=10)
                 label_frame.columnconfigure(0, weight=1)
 
                 def _command_drc(param, label_frame):
-                    delete_row_callback(self.subwrappers[param])
+                    delete_row_callback(self.subwrappers[id(param)])
                     label_frame.destroy()
 
-                self.subwrappers[p] = ParameterWrapper(
+                self.subwrappers[id(p)] = ParameterWrapper(
                     label_frame,
                     p,
                     self.callback,
@@ -302,10 +317,11 @@ class ParameterWrapper:
                     button = delete_button(label_frame_label, command=partial(_command_drc, p, label_frame))
                     button.grid(row=0, column=1, padx=10)
 
-            self.subwrappers[p].entry.grid(row=n, column=0, padx=10)
-
         self.cull_subwrappers(params)
 
+        if self.button:
+            self.button.grid(row=len(params), column=0, padx=10)
+
     def combobox_set(self, event):
         assert isinstance(self.entry, ttk.Combobox)
         if self.entry.current() != -1:
@@ -377,8 +393,11 @@ class ParameterWrapper:
             self.callback(self.parameter)
 
     def value_changed_callback(self, *_):
-        if isinstance(self.parameter, ChoiceParam) and self.entry.current() != -1:
-            self.set_choice(self.entry.current())
+        if isinstance(self.parameter, (ChoiceParam, DictChoiceParam)) and self.entry.current() != -1:
+            value = self.parameter.get_values()[self.entry.current()]
+            if value != self.parameter.value:
+                self.parameter.value = value
+                self.callback(self.parameter)
         else:
             self.var.set(self.set_value(self.var.get()))
 
@@ -411,8 +430,10 @@ class ParameterWrapper:
         self.set_value("")
 
     def destroy(self):
-        self.label.destroy()
-        self.entry.destroy()
+        if self.label:
+            self.label.destroy()
+        if self.entry:
+            self.entry.destroy()
         if self.button:
             self.button.destroy()
 
@@ -425,7 +446,7 @@ class PluginConfigurator:
         self.plugin = plugin
         self.change_callback = change_callback
 
-        self.frame = tk.Frame(tk_parent, highlightcolor="purple", highlightthickness=3)
+        self.frame = tk.Frame(tk_parent)
         self.frame.columnconfigure(0, weight=0)
         self.frame.columnconfigure(1, weight=0)
         self.frame.columnconfigure(2, weight=1)
@@ -444,13 +465,15 @@ class PluginConfigurator:
 
     def update(self) -> None:
         # If there's only a single parameter it is presented a little differently.
-        top_level = 0 if len(self.plugin.parameters) == 1 else 1
+        top_level = 0 if len(self.plugin.params) == 1 else 1
 
         # Create any new parameter wrappers needed & update existing ones
-        for n, (key, parameter) in enumerate(self.plugin.parameters.items()):
-            if key in self.wrapper_cache:
+        for n, (key, parameter) in enumerate(self.plugin.params.items()):
+            if key in self.wrapper_cache and parameter == self.wrapper_cache[key].parameter:
                 self.wrapper_cache[key].update()
             else:
+                if key in self.wrapper_cache:
+                    self.wrapper_cache[key].destroy()
                 self.wrapper_cache[key] = ParameterWrapper(
                     self.frame, parameter, self.change_parameter, level=top_level
                 )
@@ -458,6 +481,6 @@ class PluginConfigurator:
 
         # Remove any parameter wrappers no longer needed
         for key, wrapper in list(self.wrapper_cache.items()):
-            if key not in self.plugin.parameters:
+            if key not in self.plugin.params:
                 wrapper.destroy()
                 del self.wrapper_cache[key]
