@@ -1,8 +1,12 @@
+import logging
+import logging.handlers
 import multiprocessing
+import os
 import re
 import sys
 import threading
 import tkinter as tk
+from collections.abc import Callable
 from tkinter import messagebox, ttk
 from typing import Optional
 
@@ -13,16 +17,25 @@ from countess.core.config import export_config_graphviz, read_config, write_conf
 from countess.core.pipeline import PipelineGraph
 from countess.core.plugins import get_plugin_classes
 from countess.gui.config import PluginConfigurator
-from countess.gui.logger import LoggerFrame
 from countess.gui.mini_browser import MiniBrowserFrame
 from countess.gui.tabular import TabularDataFrame
 from countess.gui.tree import FlippyCanvas, GraphWrapper
-from countess.gui.widgets import ResizingFrame, ask_open_filename, ask_saveas_filename, get_icon, info_button
+from countess.gui.widgets import (
+    LabeledProgressbar,
+    ResizingFrame,
+    ask_open_filename,
+    ask_saveas_filename,
+    get_icon,
+    info_button,
+)
 from countess.utils.pandas import concat_dataframes
 
 # import faulthandler
 # faulthandler.enable(all_threads=True)
 
+logger = logging.getLogger(__name__)
+
+logging_queue: multiprocessing.Queue = multiprocessing.Queue()
 
 plugin_classes = sorted(get_plugin_classes(), key=lambda x: x.name)
 
@@ -67,9 +80,8 @@ class ConfiguratorWrapper:
     info_toplevel = None
     info_frame = None
 
-    def __init__(self, tk_parent, node, logger, change_callback):
+    def __init__(self, tk_parent, node, change_callback):
         self.node = node
-        self.logger = logger
         self.change_callback = change_callback
 
         self.frame = ResizingFrame(tk_parent, orientation=ResizingFrame.Orientation.VERTICAL, bg="darkgrey")
@@ -107,7 +119,7 @@ class ConfiguratorWrapper:
         self.config_scrollbar.grid(row=3, column=1, sticky=tk.NS)
 
         if self.node.plugin:
-            self.node.load_config(self.logger)
+            self.node.load_config()
             if self.node.notes:
                 self.show_notes_widget(self.node.notes)
             else:
@@ -220,27 +232,45 @@ class ConfiguratorWrapper:
         self.notes_widget.edit_modified(False)
 
     def config_change_callback(self, *_):
-        self.node.mark_dirty()
+        """Called immediately if a change to config has occurred."""
+        logger.debug("config_change_callback")
+
+        # Leave it a bit to see if the user is still typing, if so cancel.
         if self.config_change_task:
             self.frame.after_cancel(self.config_change_task)
-        self.config_change_task = self.frame.after(1000, self.config_change_task_callback)
+        self.config_change_task = self.frame.after(2500, self.config_change_task_callback)
 
     def config_change_task_callback(self):
-        self.config_change_task = None
+        """Called when the user makes a change then pauses for a bit"""
+        logger.debug("config_change_task_callback")
 
-        self.node_update_thread = threading.Thread(target=self.node.prerun, args=(self.logger,))
+        # if there's already an update running, delay a bit longer
+        if self.node_update_thread and self.node_update_thread.is_alive():
+            # XXX need some way to tell the thread to stop
+            logger.debug("config_change_task_callback: waiting for thread %s", self.node_update_thread)
+            self.config_change_task = self.frame.after(1000, self.config_change_task_callback)
+            return
+
+        self.config_change_task = None
+        self.node.mark_dirty()
+
+        self.node_update_thread = threading.Thread(target=self.node.prerun)
+        logger.debug("config_change_task_callback: starting thread %s", self.node_update_thread)
         self.node_update_thread.start()
 
         self.config_change_task_callback_2()
 
     def config_change_task_callback_2(self):
-        self.logger.poll()
+        logger.debug("config_change_task_callback_2")
 
-        if self.node_update_thread.is_alive():
-            self.frame.after(100, self.config_change_task_callback_2)
+        if self.node_update_thread and self.node_update_thread.is_alive():
+            logger.debug("config_change_task_callback_2: waiting for thread %s", self.node_update_thread)
+            self.frame.after(500, self.config_change_task_callback_2)
             return
 
+        logger.debug("config_change_task_callback_2: joining thread %s", self.node_update_thread)
         self.node_update_thread.join()
+        self.node_update_thread = None
 
         # XXX stop the form scrolling away when it is refreshed, by putting
         # it back where it belongs.
@@ -252,9 +282,8 @@ class ConfiguratorWrapper:
         self.config_scrollbar.set(pos1, pos2)
 
     def choose_plugin(self, plugin_class):
-        print(f"choose {self} {plugin_class}")
         self.node.plugin = plugin_class()
-        self.node.prerun(self.logger)
+        self.node.prerun()
         self.node.is_dirty = True
         self.show_config_subframe()
         self.change_callback(self.node)
@@ -290,8 +319,6 @@ class RunWindow:
 
         self.logger_frame = LoggerFrame(self.toplevel)
         self.logger_frame.grid(row=1, column=0, stick=tk.NSEW)
-        self.logger = self.logger_frame.get_logger("")
-        self.logger.info("Started")
 
         self.button = tk.Button(self.toplevel, text="Stop", command=self.on_button)
         self.button.grid(row=2, column=0, sticky=tk.EW)
@@ -302,17 +329,16 @@ class RunWindow:
         self.poll()
 
     def subproc(self):
-        self.graph.run(self.logger)
+        self.graph.run()
 
     def poll(self):
         if self.process:
             if self.process.is_alive():
                 self.logger_frame.after(1000, self.poll)
             else:
-                self.logger.info("Finished")
+                logger.info("Finished")
                 self.process = None
                 self.button["text"] = "Close"
-            self.logger_frame.poll()
 
     def on_button(self):
         if self.process:
@@ -322,14 +348,74 @@ class RunWindow:
             self.button["text"] = "Close"
             self.process = None
 
-            self.logger.info("Stopped")
-            self.logger_frame.poll()
+            logger.info("Stopped")
 
         elif self.toplevel:
             self.toplevel.destroy()
             self.toplevel = None
 
 
+class _CallbackLoggingHandler(logging.Handler):
+    """This is a tiny stub of a logging.Handler which just lets you
+    direct LogRecords to a callback instead of a handler, eg:
+    a method in some other class."""
+
+    def __init__(self, callback: Callable[[logging.LogRecord], None], *a, **k):
+        super().__init__(*a, **k)
+        self.__callback = callback
+
+    def emit(self, record: logging.LogRecord) -> None:
+        self.__callback(record)
+
+
+class LoggerFrame(tk.Frame):
+    def __init__(self, *a, **k) -> None:
+        super().__init__(*a, **k)
+        self.count = 0
+        self.text = tk.Text(self)
+        self.text.grid(sticky=tk.NSEW)
+        self.progress_bars: dict[str, LabeledProgressbar] = {}
+
+        # Start a QueueListener in its own thread,
+        # logging_callback gets called for each record
+        # received.
+        logging.handlers.QueueListener(logging_queue, _CallbackLoggingHandler(self.logging_callback)).start()
+
+        self.columnconfigure(0, weight=1)
+        self.rowconfigure(0, weight=1)
+        # self.rowconfigure(1, weight=0)
+
+        self.pbars_done: list[LabeledProgressbar] = []
+        self.hide_event()
+
+    def logging_callback(self, record: logging.LogRecord) -> None:
+        message = record.getMessage()
+        if m := re.match(r"(.*): (\d+)(%|/\d+)", message):
+            name, n1, n2 = m.groups()
+            try:
+                pbar = self.progress_bars[name]
+            except KeyError:
+                pbar = LabeledProgressbar(self, value=0)
+                self.progress_bars[name] = pbar
+                pbar.grid(sticky=tk.EW, row=len(self.progress_bars), column=0)
+            if n2 == "%":
+                pbar.progress_update(message, int(n1))
+            else:
+                pbar.progress_update(message)
+        else:
+            self.text.insert(tk.END, message + "\n")
+            self.count += 1
+
+    def hide_event(self):
+        for pbar in self.pbars_done:
+            pbar.grid_forget()
+        self.pbars_done = []
+        for pbar in self.progress_bars.values():
+            if pbar.done:
+                self.pbars_done.append(pbar)
+        self.after(5000, self.hide_event)
+
+
 class MainWindow:
     """Arrange the parts of the main window, with the FrameWrapper on the top
     left and the ConfiguratorWrapper either beside it or under it depending on
@@ -339,6 +425,7 @@ class MainWindow:
     preview_frame = None
     config_wrapper = None
     config_changed = False
+    config_filename = None
 
     def __init__(self, tk_parent: tk.Widget, config_filename: Optional[str] = None):
         self.tk_parent = tk_parent
@@ -377,8 +464,9 @@ class MainWindow:
         self.frame.add_child(self.main_subframe, weight=4)
 
         self.logger_subframe = LoggerFrame(self.frame)
-        self.logger = self.logger_subframe.get_logger("")
-        self.logger_subframe_show_task()
+        # self.logger = self.logger_subframe.get_logger("")
+        # self.logger_subframe_show_task()
+        self.frame.add_child(self.logger_subframe)
 
         self.tree_canvas = FlippyCanvas(self.main_subframe, bg="skyblue")
         self.main_subframe.add_child(self.tree_canvas)
@@ -408,6 +496,7 @@ class MainWindow:
                 return
         self.config_changed = False
         self.config_filename = None
+        self.update_title()
         if self.graph_wrapper:
             self.graph_wrapper.destroy()
         self.graph = PipelineGraph()
@@ -420,9 +509,12 @@ class MainWindow:
         if not filename:
             return
         self.config_filename = filename
+        self.config_changed = False
+        self.update_title()
+
         if self.graph_wrapper:
             self.graph_wrapper.destroy()
-        self.graph = read_config(filename)
+        self.graph = read_config([filename])
         self.graph_wrapper = GraphWrapper(self.tree_canvas, self.graph, self.node_select)
         self.node_select(None)
 
@@ -435,7 +527,9 @@ class MainWindow:
         if not filename:
             return
         write_config(self.graph, filename)
+        self.config_filename = filename
         self.config_changed = False
+        self.update_title()
         # Names may have changed on save so redraw the tree
         self.graph_wrapper.destroy()
         self.graph_wrapper = GraphWrapper(self.tree_canvas, self.graph, self.node_select)
@@ -465,7 +559,7 @@ class MainWindow:
 
     def node_select(self, node):
         if node:
-            new_config_wrapper = ConfiguratorWrapper(self.main_subframe, node, self.logger, self.node_changed)
+            new_config_wrapper = ConfiguratorWrapper(self.main_subframe, node, self.node_changed)
             if self.config_wrapper:
                 self.main_subframe.replace_child(self.config_wrapper.frame, new_config_wrapper.frame)
                 self.config_wrapper.destroy()
@@ -479,8 +573,20 @@ class MainWindow:
             self.config_wrapper = None
 
     def node_changed(self, node):
-        self.config_changed = True
         self.graph_wrapper.node_changed(node)
+        self.config_changed = True
+        self.update_title()
+
+    def update_title(self):
+        if self.config_filename:
+            title = "CountESS: " + os.path.relpath(self.config_filename)
+        else:
+            title = "CountESS " + VERSION
+
+        if self.config_changed:
+            title += " *"
+
+        self.tk_parent.title(title)
 
 
 class SplashScreen:
@@ -513,7 +619,6 @@ def make_root():
         root = tk.Tk()
         # XXX some kind of ttk style setup goes here as a fallback
 
-    root.title(f"CountESS {VERSION}")
     root.rowconfigure(0, weight=0)
     root.rowconfigure(1, weight=1)
     root.columnconfigure(0, weight=1)
@@ -529,6 +634,12 @@ def make_root():
 
 
 def main():
+    # set up a multiprocessing-compatible logging queue to bring all logging
+    # messages back to the main process.
+    logging.getLogger().addHandler(logging.handlers.QueueHandler(logging_queue))
+    logging.getLogger().addHandler(logging.StreamHandler())
+    logging.getLogger().setLevel(logging.INFO)
+
     root = make_root()
     SplashScreen(root)
     MainWindow(root, sys.argv[1] if len(sys.argv) > 1 else None)
