@@ -16,12 +16,21 @@ from rapidfuzz.distance.Levenshtein import opcodes as levenshtein_opcodes
 MIN_SEARCH_LENGTH = 10
 
 
+class TooManyVariationsException(ValueError):
+    pass
+
+
 def translate_aa(aa_seq: str) -> str:
     """translate a sequence of single-letter amino acid codes
     to a sequence of three-letter amindo acid codes
 
     >>> translate_aa("HYPERSENSITIVITIES")
     'HisTyrProGluArgSerGluAsnSerIleThrIleValIleThrIleGluSer'
+
+    >>> translate_aa("SYZYGY")
+    Traceback (most recent call last):
+        ...
+    ValueError: Invalid AA Sequence
     """
 
     try:
@@ -200,6 +209,13 @@ def find_variant_dna(ref_seq: str, var_seq: str, offset: int = 0) -> Iterable[st
     >>> list(find_variant_dna("AAACCCTTT", "AAAGGGTTT"))
     ['4_6inv']
 
+    OFFSETS
+
+    >>> list(find_variant_dna("AGAAGTAGAGG", "ATAAGAAGAGG", 100))
+    ['102G>T', '106T>A']
+
+    >>> list(find_variant_dna("AGAAGTAGAGG", "ATAAGAAGAGG", -200))
+    ['198G>T', '194T>A']
     """
 
     ref_seq = ref_seq.strip().upper()
@@ -268,9 +284,9 @@ def find_variant_dna(ref_seq: str, var_seq: str, offset: int = 0) -> Iterable[st
             assert dest_seq == ""
             # 'delete' opcode maps to HGVS 'del' operation
             if len(src_seq) == 1:
-                yield f"{start+1}del"
+                yield f"{abs(start+1)}del"
             else:
-                yield f"{start+1}_{end}del"
+                yield f"{abs(start+1)}_{abs(end)}del"
 
         elif opcode.tag == "insert":
             assert src_seq == ""
@@ -280,12 +296,12 @@ def find_variant_dna(ref_seq: str, var_seq: str, offset: int = 0) -> Iterable[st
                 # This is a duplication of one or more symbols immediately
                 # preceding this point.
                 if len(dest_seq) == 1:
-                    yield f"{start}dup"
+                    yield f"{abs(start)}dup"
                 else:
-                    yield f"{start - len(dest_seq) + 1}_{start}dup"
+                    yield f"{abs(start - len(dest_seq) + 1)}_{abs(start)}dup"
             else:
                 inserted_sequence = search_for_sequence(ref_seq, dest_seq)
-                yield f"{start}_{start+1}ins{inserted_sequence}"
+                yield f"{abs(start)}_{abs(start+1)}ins{inserted_sequence}"
 
         elif opcode.tag == "replace":
             # 'replace' opcode maps to either an HGVS '>' (single substitution) or
@@ -296,12 +312,12 @@ def find_variant_dna(ref_seq: str, var_seq: str, offset: int = 0) -> Iterable[st
             # as this code has no concept of amino acid alignment.
 
             if len(src_seq) == 1 and len(dest_seq) == 1:
-                yield f"{start+1}{src_seq}>{dest_seq}"
+                yield f"{abs(start+1)}{src_seq}>{dest_seq}"
             elif len(src_seq) == len(dest_seq) and dest_seq == reverse_complement(src_seq):
-                yield f"{start+1}_{end}inv"
+                yield f"{abs(start+1)}_{abs(end)}inv"
             else:
                 inserted_sequence = search_for_sequence(ref_seq, dest_seq)
-                yield f"{start+1}_{end}delins{inserted_sequence}"
+                yield f"{abs(start+1)}_{abs(end)}delins{inserted_sequence}"
 
 
 def find_variant_protein(ref_seq: str, var_seq: str, offset: int = 0):
@@ -323,11 +339,21 @@ def find_variant_protein(ref_seq: str, var_seq: str, offset: int = 0):
     >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGGTTCA"))
     ['Val2del']
 
+    a double AA deletion:
+
+    >>> list(find_variant_protein("ATGGTTGGTTCAGGC", "ATGTCAGGC"))
+    ['Val2_Gly3del']
+
     a single AA duplication:
 
     >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGTGGTTCA"))
     ['Gly3dup']
 
+    a double AA duplication:
+
+    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGTGTTGGTTCA"))
+    ['Val2_Gly3dup']
+
     a single AA insertion
     >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGTAAATCA"))
     ['Gly3_Ser4insLys']
@@ -364,17 +390,16 @@ def find_variant_protein(ref_seq: str, var_seq: str, offset: int = 0):
 
     >>> list(find_variant_protein("ATGGCCCCCAAATAA", "ATGGCGCCAAATTAA"))
     ['Ala2_Pro3=', 'Lys4Asn']
-
     """
 
     ref_seq = ref_seq.strip().upper()
     var_seq = var_seq.strip().upper()
 
     if not re.match("[AGTCN]+$", ref_seq):
-        raise ValueError("Invalid reference sequence")
+        raise ValueError("Invalid reference sequence")  # pragma: no cover
 
     if not re.match("[AGTCN]+$", var_seq):
-        raise ValueError("Invalid variant sequence")
+        raise ValueError("Invalid variant sequence")  # pragma: no cover
 
     frame = (3 - offset) % 3
     ref_pro = translate_dna(ref_seq[frame:])[0]
@@ -415,7 +440,7 @@ def find_variant_protein(ref_seq: str, var_seq: str, offset: int = 0):
                 if len(dest_pro) == 1:
                     yield f"{_ref(start-1)}dup"
                 else:
-                    yield f"{_ref(start-len(dest_pro))}_{_ref(start)}dup"
+                    yield f"{_ref(start-len(dest_pro))}_{_ref(start-1)}dup"
             elif start == len(ref_pro):
                 # 'extension', not quite standards compliant
                 yield f"{_ref(start-1)}ext{translate_aa(dest_pro)}"
@@ -458,7 +483,12 @@ def find_variant_protein(ref_seq: str, var_seq: str, offset: int = 0):
 
 
 def find_variant_string(
-    prefix: str, ref_seq: str, var_seq: str, max_mutations: Optional[int] = None, offset: int = 0
+    prefix: str,
+    ref_seq: str,
+    var_seq: str,
+    max_mutations: Optional[int] = None,
+    offset: int = 0,
+    minus_strand: bool = False,
 ) -> str:
     """As above, but returns a single string instead of a generator
 
@@ -497,12 +527,19 @@ def find_variant_string(
     >>> find_variant_string("p.", "ATGGTTGGTTCA", "ATGGCTGCTTCA")
     'p.Val2_Gly3delinsAlaAla'
 
+    MINUS STRAND
+
+    this example is actually comparing TGTAATC and TCTGAAC ...
+
+    >>> find_variant_string("g.", "GATTACA", "GTTCAGA", minus_strand=True)
+    'g.[2G>C;3_4insG;6del]'
+
     CHECK FOR INVALID INPUTS
 
     >>> find_variant_string("x.", "CAT", "CAT")
     Traceback (most recent call last):
      ...
-    ValueError: Only prefix types 'g.', 'n.' and 'p.' accepted at this time
+    ValueError: Only prefix types 'g.', 'c.' and 'p.' accepted at this time
 
     >>> find_variant_string("g.", "HELLO", "CAT")
     Traceback (most recent call last):
@@ -519,21 +556,25 @@ def find_variant_string(
     >>> find_variant_string("g.", "ATTACC", "GATTACA",1)
     Traceback (most recent call last):
      ...
-    ValueError: Too many variations (2) in GATTACA
+    countess.utils.variant.TooManyVariationsException: Too many variations (2) in GATTACA
     """
 
-    if prefix.endswith("g.") and not prefix.endswith("n."):
+    if minus_strand:
+        ref_seq = reverse_complement(ref_seq)
+        var_seq = reverse_complement(var_seq)
+
+    if prefix.endswith("g.") or prefix.endswith("c."):
         variations = list(find_variant_dna(ref_seq, var_seq, offset))
     elif prefix.endswith("p."):
         variations = list(find_variant_protein(ref_seq, var_seq, offset))
     else:
-        raise ValueError("Only prefix types 'g.', 'n.' and 'p.' accepted at this time")
+        raise ValueError("Only prefix types 'g.', 'c.' and 'p.' accepted at this time")
 
     if len(variations) == 0:
         return prefix + "="
 
     if max_mutations is not None and len(variations) > max_mutations:
-        raise ValueError("Too many variations (%d) in %s" % (len(variations), var_seq))
+        raise TooManyVariationsException("Too many variations (%d) in %s" % (len(variations), var_seq))
 
     if len(variations) == 1:
         return prefix + variations[0]
